package  rn.operator;

import 	 rn.vector.*;
import 	 jam.vector.*;
import 	 jam.operator.*;
import juice.operator.ZeroBiLinOp;

/**
 * <BR><!--#include file="./source.inc"--><BR>
 *
 * @author    Matt Schwab 
 */
public class MatrixVectorMul implements hasAdjoint, hasHessian
{
  /** domain of the forward operator (and range of the adjoint) */
  protected Space           input;
  /** range  of the forward operator (and domain of the forward) */
  protected Space           output;
  /** true/false as the operator is the forward/adjoint */
  protected boolean         isFwd = true;
  /**
   * Pointer to the (relative) adjoint of this operator:<BR>
   * this.isFwd == true  ---> adj.isFwd == false<BR>
   * this.isFwd == false ---> adj.isFwd == true
   */
  protected hasAdjoint              adj  = null;
  protected BiLinearOperatorAdjoint hess = null;

  private Matrix mat; 
  int     col, row; 

  /** should possibly private */
  public MatrixVectorMul(int row, int col) {
    this(new RnSpc(col), new RnSpc(row)); 
  }
  public MatrixVectorMul(RnSpc fwdDomain, RnSpc fwdRange) {
    setInput  (fwdDomain);
    setOutput (fwdRange );
    col = fwdDomain.getNSamples();
    row = fwdRange .getNSamples();
    mat = new Matrix(row, col);
  }
  public MatrixVectorMul(RnSpc domainAndRange) {
    this(domainAndRange, domainAndRange);
  }
  public MatrixVectorMul(Matrix mat) {
    this(mat.row, mat.col);
    this.mat = mat;
  }

  /** copy constructor (deep copy for coeficients; shallow copy for spaces) */
  public MatrixVectorMul(MatrixVectorMul src) {
    this.input  = (RnSpc) src.input.copy();
    this.output = (RnSpc) src.output.copy();
    this.isFwd  = src.isFwd;
    this.mat    = src.mat.copy(); 
  }
  public jam.JamObject copy() { 
    return (jam.JamObject) new MatrixVectorMul(this); 
  }

  /**
   * Apply the operator:<BR><BR>
   *
   * Depending on if the operator is the forward or adjoint:<BR>
   * forward: out = A(in)<BR>
   * adjoint: in  = A(out)<BR>
   *
   * @param add Performs y += Ax instead of y = Ax if true (where {x,y} = {in,out} -- respectively if and only if isFwd == true)
   * @param in  Vector in the domain of the forward (altered if isFwd == false)
   * @param out Vector in the range  of the forward (altered if isFwd == true)
   */

  protected void  setInput  (Space input   ) { this.input  = input ; }
  protected void  setOutput (Space output  ) { this.output = output; }
  protected void  setForward(boolean isFwd ) { this.isFwd  = isFwd;  }
  protected void  setAdjoint(hasAdjoint adj) { this.adj    = adj;    }

  protected void  createAdjoint() {
    try {
      adj = (hasAdjoint) this.clone();
    }
    catch (Exception e) {
      System.err.println(e);
    }
    ((MatrixVectorMul) adj).setForward(!this.isFwd);
    ((MatrixVectorMul) adj).setAdjoint(this);
  }

  public boolean equals(MatrixVectorMul slo2) {
    if (! this.input .equals(slo2.input )) return false; 
    if (! this.output.equals(slo2.output)) return false; 
    if (  this.isFwd != slo2.isFwd       ) return false; 
    if (! this.mat.equals(slo2.mat      )) return false; 
    return true; 
  }

  // ----------------------------------------------------------------------
  // ---------- The remaining methods are required by the superclass ------
  // ----------------------------------------------------------------------

  public Space   getDomain () { return isFwd ? input:output; }
  public Space   getRange  () { return isFwd ? output:input; }

  public Vector  image(Vector d) {
    Vector r = getRange().newMember();
    image(d, r);
    return r; 
  }

  public void    image(Vector d, Vector r) {
    if (isFwd) apply(false, d, r);
    else       apply(false, r, d);
    }
  
  public void addImage(Vector d, Vector r) {
    if (isFwd) apply(true , d, r);
    else       apply(true , r, d); 
  }

  public void residual(Vector d, Vector b, Vector r) {
    r.neg(b);
    addImage(d, r); 
  }

  public hasAdjoint     adjoint() {
    if (adj == null) createAdjoint();
    return adj;
  }

  public LinearOperator           getGradientAt  (Vector x) {
    return adjoint();
  }
  public hasAdjoint               getDerivativeAt(Vector x) {
    return this;
  }
  public BiLinearOperatorAdjoint  getHessianAt   (Vector x) {
    if (hess == null) createHessian();
    return hess;
  }
  protected void createHessian() {
    hess = new ZeroBiLinOp(getDomain(), getDomain(), getRange());
  }

  void  apply     (boolean add, Vector in, Vector out) {
    float[][] d = mat.d; 
    float[]  ii = ((RnVec)  in).d;
    float[]  oo = ((RnVec) out).d;
System.out.println("matrix: " + mat); 
System.out.println("in    : " + in );
System.out.println("out   : " + out);
    if (ii.length != col) System.err.println("Error: col dimension wrong");
    if (oo.length != row) System.err.println("Error: row dimension wrong");
    if (!add) {
      if (isFwd) out.zero(); 
      else        in.zero();
    }
    for (int i1=0; i1<row; i1++) 
    for (int i0=0; i0<col; i0++) 
      if (isFwd) oo[i1] += d[i1][i0] * ii[i0]; 
      else       ii[i0] += d[i1][i0] * oo[i1]; 
  }  
}
